@@ -26,7 +26,9 @@
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif
-#include <time.h>
+#ifdef HAVE_SYS_TYPES_H
+# include <sys/types.h>
+#endif
 
 #include "connection.h"
 #include "refbuf.h"
@@ -38,6 +40,7 @@
 
 #include "format_ogg.h"
 #include "format_mp3.h"
+#include "format_ebml.h"
 
 #include "logging.h"
 #include "stats.h"
@@ -52,15 +55,31 @@
 static int format_prepare_headers (source_t *source, client_t *client);
 
 
-format_type_t format_get_type(char *contenttype)
+format_type_t format_get_type (const char *contenttype)
 {
     if(strcmp(contenttype, "application/x-ogg") == 0)
         return FORMAT_TYPE_OGG; /* Backwards compatibility */
     else if(strcmp(contenttype, "application/ogg") == 0)
         return FORMAT_TYPE_OGG; /* Now blessed by IANA */
+    else if(strcmp(contenttype, "audio/ogg") == 0)
+        return FORMAT_TYPE_OGG;
+    else if(strcmp(contenttype, "video/ogg") == 0)
+        return FORMAT_TYPE_OGG;
+    else if(strcmp(contenttype, "audio/webm") == 0)
+        return FORMAT_TYPE_EBML;
+    else if(strcmp(contenttype, "video/webm") == 0)
+        return FORMAT_TYPE_EBML;
+    else if(strcmp(contenttype, "audio/x-matroska") == 0)
+        return FORMAT_TYPE_EBML;
+    else if(strcmp(contenttype, "video/x-matroska") == 0)
+        return FORMAT_TYPE_EBML;
+    else if(strcmp(contenttype, "video/x-matroska-3d") == 0)
+        return FORMAT_TYPE_EBML;
     else
         /* We default to the Generic format handler, which
-           can handle many more formats than just mp3 */
+           can handle many more formats than just mp3.
+	   Let's warn that this is not well supported */
+	WARN1("Unsupported or legacy stream type: \"%s\". Falling back to generic minimal handler for best effort.", contenttype);
         return FORMAT_TYPE_GENERIC;
 }
 
@@ -72,6 +91,9 @@ int format_get_plugin(format_type_t type, source_t *source)
     case FORMAT_TYPE_OGG:
         ret = format_ogg_get_plugin (source);
         break;
+    case FORMAT_TYPE_EBML:
+        ret = format_ebml_get_plugin (source);
+        break;
     case FORMAT_TYPE_GENERIC:
         ret = format_mp3_get_plugin (source);
         break;
@@ -93,15 +115,17 @@ static void find_client_start (source_t *source, client_t *client)
 {
     refbuf_t *refbuf = source->burst_point;
 
-    /* we only want to attempt a burst at connection time, not midstream */
-    if (client->intro_offset == -1)
+    /* we only want to attempt a burst at connection time, not midstream
+     * however streams like theora may not have the most recent page marked as
+     * a starting point, so look for one from the burst point */
+    if (client->intro_offset == -1 && source->stream_data_tail
+            && source->stream_data_tail->sync_point)
         refbuf = source->stream_data_tail;
     else
     {
-        long size = 0;
+        size_t size = client->intro_offset;
         refbuf = source->burst_point;
-        size = client->intro_offset;
-        while (size > 0 && refbuf->next)
+        while (size > 0 && refbuf && refbuf->next)
         {
             size -= refbuf->len;
             refbuf = refbuf->next;
@@ -126,7 +150,7 @@ static void find_client_start (source_t *source, client_t *client)
 static int get_file_data (FILE *intro, client_t *client)
 {
     refbuf_t *refbuf = client->refbuf;
-    int bytes;
+    size_t bytes;
 
     if (intro == NULL || fseek (intro, client->intro_offset, SEEK_SET) < 0)
         return 0;
@@ -134,7 +158,7 @@ static int get_file_data (FILE *intro, client_t *client)
     if (bytes == 0)
         return 0;
 
-    refbuf->len = bytes;
+    refbuf->len = (unsigned int)bytes;
     return 1;
 }
 
@@ -199,13 +223,16 @@ int format_check_http_buffer (source_t *source, client_t *client)
     {
         DEBUG0("processing pending client headers");
 
-        client->respcode = 200;
         if (format_prepare_headers (source, client) < 0)
         {
             ERROR0 ("internal problem, dropping client");
             client->con->error = 1;
             return -1;
         }
+        client->respcode = 200;
+        stats_event_inc (NULL, "listeners");
+        stats_event_inc (NULL, "listener_connections");
+        stats_event_inc (source->mount, "listener_connections");
     }
 
     if (client->pos == refbuf->len)
@@ -272,8 +299,7 @@ static int format_prepare_headers (source_t *source, client_t *client)
     ptr = client->refbuf->data;
     client->respcode = 200;
 
-    bytes = snprintf (ptr, remaining, "HTTP/1.0 200 OK\r\n"
-            "Content-Type: %s\r\n", source->format->contenttype);
+    bytes = util_http_build_header (ptr, remaining, 0, 0, 200, NULL, source->format->contenttype, NULL, NULL);
 
     remaining -= bytes;
     ptr += bytes;
@@ -309,7 +335,22 @@ static int format_prepare_headers (source_t *source, client_t *client)
             if (strcasecmp(var->name, "ice-password") &&
                 strcasecmp(var->name, "icy-metaint"))
             {
-                if (!strncasecmp("ice-", var->name, 4))
+		if (!strcasecmp(var->name, "ice-name"))
+		{
+		    ice_config_t *config;
+		    mount_proxy *mountinfo;
+
+		    config = config_get_config();
+		    mountinfo = config_find_mount (config, source->mount, MOUNT_TYPE_NORMAL);
+
+		    if (mountinfo && mountinfo->stream_name)
+		        bytes = snprintf (ptr, remaining, "icy-name:%s\r\n", mountinfo->stream_name);
+                    else
+		        bytes = snprintf (ptr, remaining, "icy-name:%s\r\n", var->value);
+
+                    config_release_config();
+		}
+                else if (!strncasecmp("ice-", var->name, 4))
                 {
                     if (!strcasecmp("ice-public", var->name))
                         bytes = snprintf (ptr, remaining, "icy-pub:%s\r\n", var->value);
@@ -336,10 +377,6 @@ static int format_prepare_headers (source_t *source, client_t *client)
     }
     avl_tree_unlock(source->parser->vars);
 
-    bytes = snprintf (ptr, remaining, "Server: %s\r\n", ICECAST_VERSION_STRING);
-    remaining -= bytes;
-    ptr += bytes;
-
     bytes = snprintf (ptr, remaining, "\r\n");
     remaining -= bytes;
     ptr += bytes;
