@@ -6,8 +6,10 @@
  * Copyright 2000-2004, Jack Moffitt <jack@xiph.org, 
  *                      Michael Smith <msmith@xiph.org>,
  *                      oddsock <oddsock@xiph.org>,
- *                      Karl Heyes <karl@xiph.org>
+ *                      Karl Heyes <karl@xiph.org>,
  *                      and others (see AUTHORS for details).
+ * Copyright 2011-2012, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
+ * Copyright 2014,      Thomas B. Ruecker <thomas@ruecker.fi>.
  */
 
 /* -*- c-basic-offset: 4; indent-tabs-mode: nil; -*- */
@@ -17,6 +19,16 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <errno.h>
+
+#ifdef WIN32
+#define _WIN32_WINNT 0x0400
+/* For getpid() */
+#include <process.h>
+#include <windows.h>
+#define snprintf _snprintf
+#define getpid _getpid
+#endif
 
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
@@ -31,11 +43,14 @@
 #include "net/resolver.h"
 #include "httpp/httpp.h"
 
-#ifdef CHUID
+#if HAVE_SYS_TYPES_H
 #include <sys/types.h>
+#endif
+#if HAVE_GRP_H
 #include <grp.h>
+#endif
+#if HAVE_PWD_H
 #include <pwd.h>
-#include <errno.h>
 #endif
 
 #include "cfgfile.h"
@@ -56,23 +71,17 @@
 
 #include <libxml/xmlmemory.h>
 
-#ifdef _WIN32
-/* For getpid() */
-#include <process.h>
-#include <windows.h>
-#define snprintf _snprintf
-#define getpid _getpid
-#endif
-
 #undef CATMODULE
 #define CATMODULE "main"
 
 static int background;
 static char *pidfile = NULL;
 
-static void _fatal_error(char *perr)
+static void _fatal_error(const char *perr)
 {
-#ifdef WIN32
+#ifdef WIN32_SERVICE
+    MessageBox(NULL, perr, "Error", MB_SERVICE_NOTIFICATION);
+#elif defined(WIN32)
     MessageBox(NULL, perr, "Error", MB_OK);
 #else
     fprintf(stdout, "%s\n", perr);
@@ -82,11 +91,12 @@ static void _fatal_error(char *perr)
 static void _print_usage(void)
 {
     printf("%s\n\n", ICECAST_VERSION_STRING);
-    printf("usage: icecast [-b -v] -c <file>\n");
+    printf("usage: icecast [-b] -c <file>\n");
+    printf("or   : icecast {-v|--version}\n");
     printf("options:\n");
-    printf("\t-c <file>\tSpecify configuration file\n");
-    printf("\t-v\t\tDisplay version info\n");
-    printf("\t-b\t\tRun icecast in the background\n");
+    printf("\t-c <file>       Specify configuration file\n");
+    printf("\t-v or --version Display version info\n");
+    printf("\t-b              Run icecast in the background\n");
     printf("\n");
 }
 
@@ -97,7 +107,7 @@ static void _stop_logging(void)
     log_close(playlistlog);
 }
 
-static void _initialize_subsystems(void)
+void initialize_subsystems(void)
 {
     log_initialize();
     thread_initialize();
@@ -107,23 +117,16 @@ static void _initialize_subsystems(void)
     connection_initialize();
     global_initialize();
     refbuf_initialize();
-#if !defined(WIN32) || defined(WIN32_SERVICE)
-    /* win32 GUI needs to do the initialise before here */
+
     xslt_initialize();
 #ifdef HAVE_CURL_GLOBAL_INIT
     curl_global_init (CURL_GLOBAL_ALL);
 #endif
-#endif
 }
 
-static void _shutdown_subsystems(void)
+void shutdown_subsystems(void)
 {
     fserve_shutdown();
-#if !defined(WIN32) || defined(WIN32_SERVICE)
-    /* If we do the following cleanup on the win32 GUI then the app will crash when libxml2 is
-     * initialised again as the process doesn't terminate */
-    xslt_shutdown();
-#endif
     refbuf_shutdown();
     slave_shutdown();
     auth_shutdown();
@@ -144,6 +147,7 @@ static void _shutdown_subsystems(void)
     /* Now that these are done, we can stop the loggers. */
     _stop_logging();
     log_shutdown();
+    xslt_shutdown();
 }
 
 static int _parse_config_opts(int argc, char **argv, char *filename, int size)
@@ -173,7 +177,7 @@ static int _parse_config_opts(int argc, char **argv, char *filename, int size)
             background = 1;
 #endif
         }
-        if (strcmp(argv[i], "-v") == 0) {
+        if (strcmp(argv[i], "-v") == 0 || strcmp(argv[i], "--version") == 0) {
             fprintf(stdout, "%s\n", ICECAST_VERSION_STRING);
             exit(0);
         }
@@ -285,7 +289,7 @@ static int _start_listening(void)
         if (sock_listen(global.serversock[i], ICE_LISTEN_QUEUE) == SOCK_ERROR)
             return 0;
 
-        sock_set_blocking(global.serversock[i], SOCK_NONBLOCK);
+        sock_set_blocking(global.serversock[i], 0);
     }
 
     return 1;
@@ -339,7 +343,7 @@ static void _server_proc(void)
 static void _ch_root_uid_setup(void)
 {
    ice_config_t *conf = config_get_config_unlocked();
-#ifdef CHUID
+#ifdef HAVE_SETUID
    struct passwd *user;
    struct group *group;
    uid_t uid=-1;
@@ -365,7 +369,7 @@ static void _ch_root_uid_setup(void)
    }
 #endif
 
-#ifdef CHROOT
+#if HAVE_CHROOT
    if (conf->chroot)
    {
        if(getuid()) /* root check */
@@ -383,7 +387,7 @@ static void _ch_root_uid_setup(void)
 
    }   
 #endif
-#ifdef CHUID
+#if HAVE_SETUID
 
    if(conf->chuid)
    {
@@ -393,14 +397,15 @@ static void _ch_root_uid_setup(void)
            return;
        }
 
-       if(gid != -1) {
+       if(uid != (uid_t)-1 && gid != (gid_t)-1) {
            if(!setgid(gid))
                fprintf(stdout, "Changed groupid to %i.\n", (int)gid);
            else
                fprintf(stdout, "Error changing groupid: %s.\n", strerror(errno));
-       }
-
-       if(uid != -1) {
+           if(!initgroups(conf->user, gid))
+               fprintf(stdout, "Changed supplementary groups based on user: %s.\n", conf->user);
+	   else
+               fprintf(stdout, "Error changing supplementary groups: %s.\n", strerror(errno));
            if(!setuid(uid))
                fprintf(stdout, "Changed userid to %i.\n", (int)uid);
            else
@@ -425,9 +430,10 @@ int main(int argc, char **argv)
     */
     res = _parse_config_opts(argc, argv, filename, 512);
     if (res == 1) {
+#if !defined(_WIN32) || defined(_CONSOLE) || defined(__MINGW32__) || defined(__MINGW64__)
         /* startup all the modules */
-        _initialize_subsystems();
-
+        initialize_subsystems();
+#endif
         /* parse the config file */
         config_get_config();
         ret = config_initial_parse_file(filename);
@@ -451,7 +457,9 @@ int main(int argc, char **argv)
                 _fatal_error("XML config parsing error");
                 break;
             }
-            _shutdown_subsystems();
+#if !defined(_WIN32) || defined(_CONSOLE) || defined(__MINGW32__) || defined(__MINGW64__)
+            shutdown_subsystems();
+#endif
             return 1;
         }
     } else if (res == -1) {
@@ -465,7 +473,7 @@ int main(int argc, char **argv)
     /* Bind socket, before we change userid */
     if(!_server_proc_init()) {
         _fatal_error("Server startup failed. Exiting");
-        _shutdown_subsystems();
+        shutdown_subsystems();
         return 1;
     }
 
@@ -474,14 +482,14 @@ int main(int argc, char **argv)
     stats_initialize(); /* We have to do this later on because of threading */
     fserve_initialize(); /* This too */
 
-#ifdef CHUID 
+#ifdef HAVE_SETUID 
     /* We'll only have getuid() if we also have setuid(), it's reasonable to
      * assume */
     if(!getuid()) /* Running as root! Don't allow this */
     {
         fprintf(stderr, "ERROR: You should not run icecast2 as root\n");
         fprintf(stderr, "Use the changeowner directive in the config file\n");
-        _shutdown_subsystems();
+        shutdown_subsystems();
         return 1;
     }
 #endif
@@ -491,7 +499,7 @@ int main(int argc, char **argv)
 
     if (!_start_logging()) {
         _fatal_error("FATAL: Could not start logging");
-        _shutdown_subsystems();
+        shutdown_subsystems();
         return 1;
     }
 
@@ -512,9 +520,9 @@ int main(int argc, char **argv)
     _server_proc();
 
     INFO0("Shutting down");
-
-    _shutdown_subsystems();
-
+#if !defined(_WIN32) || defined(_CONSOLE) || defined(__MINGW32__) || defined(__MINGW64__)
+    shutdown_subsystems();
+#endif
     if (pidfile)
     {
         remove (pidfile);
